/*
 * crypto/testing/digesttest.c
 *
 * Written by Andrew McDonald <andrew@mcdonald.org.uk> 1999-11-16
 * with bits from speed.c:
 * written by Alexander Kjeldaas <astor@guardian.no> 1998-11-23
 *
 * Copyright 1999 by Andrew McDonald.
 * Some code, Copyright 1998 by Alexander Kjeldaas.
 * Redistribution of this file is permitted under the GNU Public License.
 */

#define __KERNEL__

#define printk printk_dummy

#include <linux/crypto.h>

#undef memcpy
#undef memset
#undef printk

#include "libcfuncs.h"
#include "misc.h"


int digest_test(struct digest_implementation *di, FILE *file);


int main(int argc, char* argv[])
{
	int digest;
	struct digest_implementation *di;
	FILE *file; 

	if (argc < 2 || argc > 3) {
		printf("Usage:\n");
		printf("digesttest digestnum [filename]\n");
		exit(1);
	}

	cryptoapi_init();
	digest = atoi(argv[1]);
	di = find_digest_by_id(digest);
	if (!di)
		exit(1);
	PRINTF("Testing digest %s, number %d...\n", di->trans.t_name,
	       di->trans.t_id);

	if (argc < 3)
		file = stdin;
	else {
		if (!(file = fopen(argv[2], "rb"))) {
			fprintf(stderr, "Can't open file %s", argv[2]);
			exit(1);
		}
	}

	digest_test(di, file);
	exit(0);
}

int digest_test(struct digest_implementation *di, FILE *file)
{
	struct digest_context* cx;
	u8* output;
	u8 buf[16384];
	int i, j;

        cx = (struct digest_context *) malloc(sizeof(struct digest_context));
        cx->di = di;
        cx->digest_info = (u32 *) malloc(di->working_size);
	output = (unsigned char *) malloc(di->blocksize);
	di->open(cx);

	while (!feof(file)) {
		i = fread(buf, 1, 16384, file);
		di->update(cx, buf, i);
	}

	di->close(cx, output);

	fclose(file);

	for (i = 0; i < (di->blocksize / sizeof(unsigned int)); i++) {
		for (j = 0; j < sizeof(unsigned int); j++) {
			printf("%02X", output[i*4+j]);
		}
		printf(" ");
	}
	printf("\n");
	return 0;
}
