/*
 * crypto/testing/misc.c
 *
 * Written by Alexander Kjeldaas <astor@guardian.no> 2000-10-08
 *
 * Copyright 2000 by Alexander Kjeldaas. Redistribution of this file
 * is permitted under the GNU Public License.
 */

/* Misc. functions */

#define __KERNEL__

#define printk printk_dummy

#include <linux/crypto.h>
#undef memcpy
#undef memset
#undef printk

#include "libcfuncs.h"
#include "misc.h"


int h2int(char c) 
{
	c = tolower(c);
	if (c >= '0' && c <= '9')
		return c - '0';
	return c - 'a' + 10;
}

char int2h(int h)
{
	if (h >= 0 && h <= 9)
		return (char)(h + '0');
	return (char)h - 10 + 'a';
}

int parsehex(char *h, char *b)
{
	int size = 0;
	/* int i; */
	char *hex = h, *buf = b;
	if ((strlen(hex) & 1) == 1) {
		*buf = h2int(*hex);
		buf++; hex++;
		size++;
	}
	for (; *hex; hex+=2, buf++, size++) {
		*buf = (h2int(hex[0]) << 4) + h2int(hex[1]);
	}
	/* net2host(b, size); */
	return size;
}

void net2host(char *x, int size)
{
	int *y = (int *)x;
	size /= sizeof(int);
	for (; size >= 0; size--)
		y[size] = ntohl(y[size]);
}

asmlinkage int printk(const char * fmt, va_list args)
{
	return vprintf(fmt, args);
}


void dump_buffer(unsigned char *buf, int size)
{
	int i;
	/* net2host(buf, size); */
	for (i = 0; i < size; i++, buf++) {
		PRINTF("%c", int2h((*buf) >> 4));
		PRINTF("%c", int2h((*buf) & 0xf));
	}
		
	PRINTF("\n");
}
	
int make_random_buf(char *buf, int size)
{
	int i;
	for (i = 0; i < size; i++) {
		buf[i] = rand();
	}
	return 0;
}

int my_memcmp(char *buf1, char *buf2, int size)
{
	int i;
	int *b1 = (int *)buf1;
	int *b2 = (int *)buf2;
	for (i = 0; i < size; i+=sizeof(int)) {
		if (*b1 != *b2) {
			PRINTF("Error at offset %d: buf1=%0x, buf2=%0x\n",
			       i, *b1, *b2);
			return 1;
		}
		b1++;b2++;
	}
	return memcmp(buf1, buf2, size);
}


void dump_u8_buffer(u8 *buf, int number)
{
	int i;

	for(i=0;i<number;i++)
	{
		PRINTF("%02x ",buf[i]);
		if(i && ((i+1)%16)==0)
			PRINTF("\n");
		else if(((i+1)%4)==0)
			PRINTF("  ");
	}
	if((i%16)!=0)
		PRINTF("\n");
}

void dump_u16_buffer(u16 *buf, int number)
{
	int i;

	for(i=0;i<number;i++)
	{
		PRINTF("%04x  ",buf[i]);
		if(i && ((i+1)%8)==0)
			PRINTF("\n");
		else if(((i+1)%2)==0)
			PRINTF("  ");
		}
	if((i%8)!=0)
		PRINTF("\n");
}

void dump_u32_buffer(u32 *buf, int number)
{
	int i;

	for(i=0;i<number;i++)
	{
		PRINTF("%08x ",buf[i]);
		if(i && ((i+1)%4)==0)
			PRINTF("\n");
		else
			PRINTF("     ");
	}
	if((i%4)!=0)
		PRINTF("\n");
}

int swab32_buf(char *buf, int size)
{
	int i;
	u32 *p = (u32 *)buf; 
	for (i = 0; i < size/sizeof(u32); i++) {
		*p = swab32(*p);
		p++;
	}
	return 0;
}

	
