/*
 *	Firewall for the ethernet bridge, using ipchains
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *	Joachim Ott			<ott@ardala.han.de>
 *
 *	$Id$
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/init.h>
#include <linux/firewall.h>
#include <linux/ip_fw.h>
#include <linux/if_bridge.h>
#include <asm/uaccess.h>
#include "br_private.h"



int brfw_forward_check(struct firewall_ops *this, int pf, struct device *dev,
		       void *phdr, void *arg, struct sk_buff **pskb)
{
	struct net_bridge *br;
	int err;
	struct iphdr *ip;
	unsigned char *ptr;
	struct sk_buff *skb;

	br = dev->br_port->br;
	skb = *pskb;
	ptr = skb->data;
	ip = (struct iphdr *)(ptr + ETH_HLEN);

	if (ptr[12] != 0x08 || ptr[13] != 0x00)		/* ETH_P_IP */
		return FW_ACCEPT;

	err = ip_fw_check_from_bridge(br->name, skb, ip, arg);
	if (err != FW_SKIP)
		return err;

	return FW_ACCEPT;
}

int brfw_input_check(struct firewall_ops *this, int pf, struct device *dev,
		     void *phdr, void *arg, struct sk_buff **pskb)
{
	return FW_ACCEPT;
}

int brfw_output_check(struct firewall_ops *this, int pf, struct device *dev,
		      void *phdr, void *arg, struct sk_buff **pskb)
{
	return FW_ACCEPT;
}




struct firewall_ops brfw_ops =
{
	NULL,
	brfw_forward_check,
	brfw_input_check,
	brfw_output_check,
	PF_BRIDGE,
	0
};

__initfunc(int brfw_init(void))
{
	printk(KERN_INFO "NET4: Ethernet Bridge Firewall 001 for NET4.0\n");

	if (register_firewall(PF_BRIDGE, &brfw_ops) < 0) {
		printk(KERN_ALERT "br: unable to register PF_BRIDGE firewall\n");
		return 1;
	}

	return 0;
}

void brfw_uninit(void)
{
	unregister_firewall(PF_BRIDGE, &brfw_ops);
}
